from django.contrib import admin
from django.contrib.auth import get_user_model

from . import models


class LineItemInline(admin.TabularInline):
    model = models.LineItem
    raw_id_fields = ("user",)
    extra = 0


@admin.register(models.Payment)
class PaymentAdmin(admin.ModelAdmin):
    date_hierarchy = "created_at"
    inlines = [LineItemInline]
    list_display = (
        "user",
        "created_at",
        "charged_at",
        "amount",
        "payment_service_provider",
        "email",
    )
    list_filter = ("charged_at", "payment_service_provider")
    raw_id_fields = ("user",)
    search_fields = (
        "email",
        "transaction",
        "user__{}".format(get_user_model().USERNAME_FIELD),
    )


@admin.register(models.LineItem)
class LineItemAdmin(admin.ModelAdmin):
    date_hierarchy = "created_at"
    list_display = ("user", "payment", "created_at", "title", "amount")
    raw_id_fields = ("user", "payment")
    search_fields = ("title", "user__{}".format(get_user_model().USERNAME_FIELD))



from django.contrib import admin
from django.contrib.auth import get_user_model

from user_payments.exceptions import UnknownPeriodicity
from . import models


class SubscriptionPeriodInline(admin.TabularInline):
    model = models.SubscriptionPeriod
    raw_id_fields = ("line_item",)
    extra = 0


@admin.register(models.Subscription)
class SubscriptionAdmin(admin.ModelAdmin):
    inlines = [SubscriptionPeriodInline]
    list_display = (
        "user",
        "title",
        "code",
        "starts_on",
        "ends_on",
        "periodicity",
        "amount",
        "paid_until",
        "renew_automatically",
    )
    list_filter = ("renew_automatically", "code")
    radio_fields = {"periodicity": admin.HORIZONTAL}
    raw_id_fields = ("user",)
    search_fields = (
        "title",
        "code",
        "user__{}".format(get_user_model().USERNAME_FIELD),
    )

    def get_inline_instances(self, request, obj=None):
        if obj is None:
            return []
        return super().get_inline_instances(request, obj=obj)

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if not change:
            try:
                obj.create_periods()
            except UnknownPeriodicity as exc:
                self.message_user(request, exc)


@admin.register(models.SubscriptionPeriod)
class SubscriptionPeriodAdmin(admin.ModelAdmin):
    list_display = ("subscription", "starts_on", "ends_on", "line_item")
    raw_id_fields = ("subscription", "line_item")
    search_fields = [
        "subscription__{}".format(field) for field in SubscriptionAdmin.search_fields
    ]



from types import SimpleNamespace

from django.apps import AppConfig
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _

import stripe


class StripeCustomersConfig(AppConfig):
    name = "user_payments.stripe_customers"
    verbose_name = capfirst(_("stripe customers"))

    def ready(self):
        from django.conf import settings

        stripe.api_key = settings.STRIPE_SECRET_KEY
        self.settings = SimpleNamespace(
            publishable_key=settings.STRIPE_PUBLISHABLE_KEY,
            secret_key=settings.STRIPE_SECRET_KEY,
        )



import json
import re

from django.contrib import admin
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _

from . import models


def sanitize_value(matchobj):
    return "%s%s" % (matchobj.group(1), "*" * len(matchobj.group(3)))


def sanitize(data, *, key=None):
    if isinstance(data, dict):
        return {key: sanitize(value, key=key) for key, value in data.items()}
    elif isinstance(data, list):
        return [sanitize(item) for item in data]
    elif key in ("fingerprint", "last4"):
        return "*" * len(data)
    elif isinstance(data, str):
        return re.sub(r"((cus_|sub_|card_)\w{6})(\w+)", sanitize_value, data)
    else:
        # Bools, ints, etc.
        return data


@admin.register(models.Customer)
class CustomerAdmin(admin.ModelAdmin):
    list_display = ("user", "customer_id_admin", "created_at", "updated_at")
    raw_id_fields = ("user",)
    search_fields = ("user__email",)

    def customer_id_admin(self, instance):
        return str(instance)

    customer_id_admin.short_description = _("customer ID")

    def customer_admin(self, instance):
        return format_html(
            "<pre>{}</pre>",
            json.dumps(sanitize(instance.customer), sort_keys=True, indent=4),
        )

    customer_admin.short_description = _("customer")

    def get_fields(self, request, obj=None):
        return (
            ("user", "created_at", "customer_id_admin", "customer_admin")
            if obj
            else ("user", "customer_id")
        )

    def get_readonly_fields(self, request, obj=None):
        return ("customer_id_admin", "customer_admin") if obj else ()



from datetime import date, datetime, time, timedelta

from django.apps import apps
from django.conf import settings
from django.db import models, transaction
from django.db.models import Max, Q, signals
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from user_payments.models import LineItem, Payment

from .utils import recurring


class SubscriptionQuerySet(models.QuerySet):
    def for_code(self, code):
        try:
            return self.get(code=code)
        except self.model.DoesNotExist:
            return None


class SubscriptionManager(models.Manager):
    def create(self, *, user, code, periodicity, amount, **kwargs):
        """
        Make it a ``TypeError`` to forget fields.
        """
        return super().create(
            user=user, code=code, periodicity=periodicity, amount=amount, **kwargs
        )

    def ensure(self, *, user, code, **kwargs):
        """
        Ensure that the user is subscribed to the subscription specified by
        ``code``. Pass additional fields for the subscription as kwargs.

        If the subscription is still in a paid period this also ensures that
        new subscription periods aren't created too early.
        """
        with transaction.atomic():
            changed = False

            try:
                subscription = self.get(user=user, code=code)
            except Subscription.DoesNotExist:
                subscription = self.create(user=user, code=code, **kwargs)
            else:
                for key, value in kwargs.items():
                    if getattr(subscription, key) != value:
                        changed = True
                        setattr(subscription, key, value)

                subscription.save()

                if not changed:
                    return subscription

            subscription.delete_pending_periods()

            if subscription.paid_until > date.today():
                # paid_until might already have been changed in the save()
                # call above. So look at paid_until and not at starts_on
                subscription.starts_on = subscription.paid_until + timedelta(days=1)

            subscription.save()
        return subscription

    def create_periods(self):
        for subscription in self.filter(renew_automatically=True):
            subscription.create_periods()

    def disable_autorenewal(self):
        """
        Disable autorenewal for subscriptions that are past due

        Uses the ``USER_PAYMENTS['disable_autorenewal_after']`` timedelta to
        determine the timespan after which autorenewal is disabled for unpaid
        subscriptions. Defaults to 15 days.
        """
        s = apps.get_app_config("user_payments").settings
        for subscription in self.filter(
            renew_automatically=True,
            paid_until__lt=timezone.now() - s.disable_autorenewal_after,
        ):
            subscription.cancel()


class Subscription(models.Model):
    """
    How to quickly generate a new subscription and fetch the first payment::

        user = request.user  # Fetch the user from somewhere

        subscription, created = Subscription.objects.get_or_create(
            user=user,
            code='plan',  # Or whatever makes sense for you
            defaults={
                'title': 'You should want to provide this',
                'periodicity': 'yearly',  # or monthly, or weekly. NOT manually.
                'amount': 60,
            },
        )
        for period in subscription.create_periods():
            period.create_line_item()
        first_payment = Payment.objects.create_pending(user=user)
    """

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="user_subscriptions",
        verbose_name=_("user"),
    )
    code = models.CharField(
        _("code"),
        max_length=20,
        help_text=_(
            "Codes must be unique per user. Allows identifying the subscription."
        ),
    )
    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    title = models.CharField(_("title"), max_length=200)
    starts_on = models.DateField(_("starts on"))
    ends_on = models.DateField(_("ends on"), blank=True, null=True)
    periodicity = models.CharField(
        _("periodicity"),
        max_length=20,
        choices=[
            ("yearly", _("yearly")),
            ("quarterly", _("quarterly")),
            ("monthly", _("monthly")),
            ("weekly", _("weekly")),
            ("manually", _("manually")),
        ],
    )
    amount = models.DecimalField(_("amount"), max_digits=10, decimal_places=2)

    renew_automatically = models.BooleanField(_("renew automatically"), default=True)
    paid_until = models.DateField(_("paid until"), blank=True)

    objects = SubscriptionManager.from_queryset(SubscriptionQuerySet)()

    class Meta:
        unique_together = (("user", "code"),)
        verbose_name = _("subscription")
        verbose_name_plural = _("subscriptions")

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        if not self.starts_on:
            self.starts_on = date.today()
        if not self.paid_until or self.paid_until < self.starts_on:
            # New subscription instance or restarted subscription with
            # inactivity period.
            self.paid_until = self.starts_on - timedelta(days=1)
        super().save(*args, **kwargs)

        # Update unbound line items with new amount.
        LineItem.objects.unbound().filter(subscriptionperiod__subscription=self).update(
            amount=self.amount
        )

    save.alters_data = True

    def update_paid_until(self, save=True):
        self.paid_until = self.periods.paid().aggregate(m=Max("ends_on"))["m"]
        if save:
            self.save()

    update_paid_until.alters_data = True

    @property
    def starts_at(self):
        return timezone.make_aware(
            datetime.combine(self.starts_on, time.min), timezone.get_default_timezone()
        )

    @property
    def ends_at(self):
        return (
            timezone.make_aware(
                datetime.combine(self.ends_on, time.max),
                timezone.get_default_timezone(),
            )
            if self.ends_on
            else None
        )

    @property
    def paid_until_at(self):
        return timezone.make_aware(
            datetime.combine(self.paid_until, time.max), timezone.get_default_timezone()
        )

    @property
    def grace_period_ends_at(self):
        s = apps.get_app_config("user_payments").settings
        return self.paid_until_at + s.grace_period

    @property
    def is_active(self):
        s = apps.get_app_config("user_payments").settings
        return timezone.now() <= self.paid_until_at + s.grace_period

    @property
    def in_grace_period(self):
        return self.paid_until_at <= timezone.now() <= self.grace_period_ends_at

    def create_periods(self, *, until=None):
        """
        Create period instances for this subscription, up to either today or
        the end of the subscription, whichever date is earlier.

        ``until`` is interpreted as "up to and including".
        """
        end = until or date.today()
        if self.ends_on:
            end = min(self.ends_on, end)
        days = recurring(self.starts_on, self.periodicity)
        this_start = next(days)

        latest_ends_on = self.periods.aggregate(m=Max("ends_on"))["m"]
        periods = []

        if this_start <= end:
            while True:
                next_start = next(days)
                if latest_ends_on is None or this_start > latest_ends_on:
                    periods.append(
                        self.periods.create(
                            starts_on=this_start, ends_on=next_start - timedelta(days=1)
                        )
                    )
                this_start = next_start

                if this_start > end:
                    break

        return periods

    create_periods.alters_data = True

    def delete_pending_periods(self):
        for period in self.periods.select_related("line_item__payment"):
            line_item = period.line_item
            if line_item:
                if line_item.payment:
                    if line_item.payment.charged_at:
                        continue
                    line_item.payment.cancel_pending()
            period.delete()
            if line_item:
                line_item.delete()

    delete_pending_periods.alters_data = True

    def cancel(self):
        self.ends_on = self.paid_until
        self.renew_automatically = False
        self.save()

        self.delete_pending_periods()

    cancel.alters_data = True


def payment_changed(sender, instance, **kwargs):
    affected = SubscriptionPeriod.objects.filter(line_item__payment=instance.pk).values(
        "subscription"
    )
    for subscription in Subscription.objects.filter(pk__in=affected):
        subscription.update_paid_until()


signals.post_save.connect(payment_changed, sender=Payment)
signals.post_delete.connect(payment_changed, sender=Payment)


class SubscriptionPeriodManager(models.Manager):
    def paid(self):
        """
        Return subscription periods that have been paid for.
        """
        return self.filter(line_item__payment__charged_at__isnull=False)

    def create_line_items(self, *, until=None):
        for period in self.filter(
            line_item__isnull=True, starts_on__lte=until or date.today()
        ):
            period.create_line_item()

    def zeroize_pending_periods(self, *, lasting_until=None):
        LineItem.objects.filter(
            id__in=self.filter(
                ~Q(id__in=self.paid()), Q(ends_on__lt=lasting_until or date.today())
            ).values("line_item")
        ).update(amount=0)


class SubscriptionPeriod(models.Model):
    subscription = models.ForeignKey(
        Subscription,
        on_delete=models.CASCADE,
        related_name="periods",
        verbose_name=_("subscription"),
    )
    starts_on = models.DateField(_("starts on"))
    ends_on = models.DateField(_("ends on"))
    line_item = models.OneToOneField(
        LineItem,
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("line item"),
    )

    objects = SubscriptionPeriodManager()

    class Meta:
        get_latest_by = "starts_on"
        unique_together = (("subscription", "starts_on"),)
        verbose_name = _("subscription period")
        verbose_name_plural = _("subscription periods")

    def __str__(self):
        return "%s (%s - %s)" % (self.subscription, self.starts_on, self.ends_on)

    def create_line_item(self):
        """
        Create a user payments line item for this subscription period.
        """
        if not self.line_item:
            self.line_item = LineItem.objects.create(
                user=self.subscription.user,
                title=str(self),
                amount=self.subscription.amount,
            )
            self.save()

    create_line_item.alters_data = True

    @property
    def starts_at(self):
        return timezone.make_aware(
            datetime.combine(self.starts_on, time.min), timezone.get_default_timezone()
        )

    @property
    def ends_at(self):
        return (
            timezone.make_aware(
                datetime.combine(self.ends_on, time.max),
                timezone.get_default_timezone(),
            )
            if self.ends_on
            else None
        )



import logging
from enum import Enum

from django.contrib.auth import get_user_model

from user_payments.models import LineItem, Payment


logger = logging.getLogger(__name__)


class Result(Enum):
    #: Processor has successfully handled the payment
    SUCCESS = 1
    #: Preconditions of the processor are not met (e.g. no credit card
    #: information). Try next processor.
    FAILURE = 2
    #: Terminates processing for this payment, do not run other processors
    TERMINATE = 3

    def __bool__(self):
        raise ResultError("Results may not be interpreted as bools")


class ResultError(Exception):
    pass


def process_payment(payment, *, processors, cancel_on_failure=True):
    logger.info(
        "Processing: %(payment)s by %(email)s",
        {"payment": payment, "email": payment.email},
    )
    success = False

    try:
        for processor in processors:
            # Success processing the payment?
            result = processor(payment)
            if result == Result.SUCCESS:
                logger.info(
                    "Success: %(payment)s by %(email)s with %(processor)s",
                    {
                        "payment": payment,
                        "email": payment.email,
                        "processor": processor.__name__,
                    },
                )
                success = True
                return True

            elif result == Result.TERMINATE:
                logger.info(
                    "Warning: Processor %(processor)s terminates processing of"
                    " %(payment)s by %(email)s",
                    {
                        "payment": payment,
                        "email": payment.email,
                        "processor": processor.__name__,
                    },
                )
                break

            elif result == Result.FAILURE:
                # It's fine, do nothing.
                pass

            else:
                raise ResultError(
                    "Invalid result %r from %s" % (result, processor.__name__)
                )

        else:
            logger.warning(
                "Warning: No success processing %(payment)s by %(email)s",
                {"payment": payment, "email": payment.email},
            )

    except Exception:
        logger.exception("Exception while processing %(payment)s by %(email)s")
        raise

    finally:
        if not success and cancel_on_failure:
            payment.cancel_pending()


def process_unbound_items(*, processors):
    for user in (
        get_user_model()
        .objects.filter(id__in=LineItem.objects.unbound().values("user"))
        .select_related("stripe_customer")
    ):
        payment = Payment.objects.create_pending(user=user)
        if payment:  # pragma: no branch (very unlikely)
            process_payment(payment, processors=processors)


def process_pending_payments(*, processors):
    for payment in Payment.objects.pending():
        process_payment(payment, cancel_on_failure=False, processors=processors)



import json

from django import http
from django.apps import apps
from django.conf.urls import url
from django.contrib import messages
from django.shortcuts import get_object_or_404, redirect
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

import stripe
from mooch.base import BaseMoocher, csrf_exempt_m, require_POST_m
from mooch.signals import post_charge

from .models import Customer


class StripeMoocher(BaseMoocher):
    identifier = "stripe"
    title = _("Pay with Stripe")

    def get_urls(self):
        return [url(r"^stripe_charge/$", self.charge_view, name="stripe_charge")]

    def payment_form(self, request, payment):
        try:
            customer = request.user.stripe_customer
        except (AttributeError, Customer.DoesNotExist):
            customer = None

        s = apps.get_app_config("stripe_customers").settings

        return render_to_string(
            "stripe_customers/payment_form.html",
            {
                "moocher": self,
                "payment": payment,
                "publishable_key": s.publishable_key,
                "customer": customer,
                "charge_url": reverse("%s:stripe_charge" % self.app_name),
                "LANGUAGE_CODE": getattr(request, "LANGUAGE_CODE", "auto"),
            },
            request=request,
        )

    @csrf_exempt_m
    @require_POST_m
    def charge_view(self, request):
        s = apps.get_app_config("user_payments").settings

        try:
            customer = request.user.stripe_customer
        except (AttributeError, Customer.DoesNotExist):
            customer = None

        instance = get_object_or_404(self.model, id=request.POST.get("id"))
        instance.payment_service_provider = self.identifier
        instance.transaction = repr(
            {key: values for key, values in request.POST.lists() if key != "token"}
        )
        instance.save()

        try:
            if (
                not customer
                and request.POST.get("token")
                and request.user.is_authenticated
            ):
                customer = Customer.objects.with_token(
                    user=request.user, token=request.POST["token"]
                )

            if customer:
                # FIXME Only with valid default source
                charge = stripe.Charge.create(
                    customer=customer.customer_id,
                    amount=instance.amount_cents,
                    currency=s.currency,
                    idempotency_key="charge-%s-%s"
                    % (instance.id.hex, instance.amount_cents),
                )
            else:
                # TODO create customer anyway, and stash away the customer ID
                # for associating with a user account after succesful payment?
                charge = stripe.Charge.create(
                    source=request.POST["token"],
                    amount=instance.amount_cents,
                    currency=s.currency,
                    idempotency_key="charge-%s-%s"
                    % (instance.id.hex, instance.amount_cents),
                )

            # TODO Error handling

            instance.charged_at = timezone.now()
            instance.transaction = json.dumps(charge)
            instance.save()

            post_charge.send(sender=self, payment=instance, request=request)

            return http.HttpResponseRedirect(self.success_url)

        except stripe.error.CardError as exc:
            messages.error(
                request, _("Card error: %s") % (exc._message or _("No details"))
            )
            return redirect(self.failure_url)



# Generated by Django 2.0.5 on 2018-05-21 05:11

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="LineItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2, max_digits=10, verbose_name="amount"
                    ),
                ),
            ],
            options={
                "verbose_name": "line item",
                "verbose_name_plural": "line items",
                "ordering": ["-created_at"],
            },
        ),
        migrations.CreateModel(
            name="Payment",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                (
                    "charged_at",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="charged at"
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2, max_digits=10, verbose_name="amount"
                    ),
                ),
                (
                    "payment_service_provider",
                    models.CharField(
                        blank=True,
                        max_length=100,
                        verbose_name="payment service provider",
                    ),
                ),
                ("email", models.EmailField(max_length=254, verbose_name="email")),
                (
                    "transaction",
                    models.TextField(blank=True, verbose_name="transaction"),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="user_payments",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "payment",
                "verbose_name_plural": "payments",
                "ordering": ("-created_at",),
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="lineitem",
            name="payment",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="lineitems",
                to="user_payments.Payment",
                verbose_name="payment",
            ),
        ),
        migrations.AddField(
            model_name="lineitem",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="user_lineitems",
                to=settings.AUTH_USER_MODEL,
                verbose_name="user",
            ),
        ),
    ]



from django.apps import AppConfig
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _


class UserSubscriptions(AppConfig):
    name = "user_payments.user_subscriptions"
    verbose_name = capfirst(_("user subscriptions"))



from datetime import timedelta
from types import SimpleNamespace

from django.apps import AppConfig
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _


class UserPayments(AppConfig):
    name = "user_payments"
    verbose_name = capfirst(_("user payments"))
    default_settings = {
        "currency": "CHF",
        "grace_period": timedelta(days=7),
        "disable_autorenewal_after": timedelta(days=15),
    }

    def ready(self):
        from django.conf import settings

        self.settings = SimpleNamespace(
            **{**self.default_settings, **getattr(settings, "USER_PAYMENTS", {})}
        )



from django.conf import settings
from django.db import models, transaction
from django.db.models import Q
from django.utils import timezone
from django.utils.translation import gettext, gettext_lazy as _

from mooch.models import Payment as AbstractPayment


class PaymentQuerySet(models.QuerySet):
    def pending(self):
        return self.filter(charged_at__isnull=True)


class PaymentManager(models.Manager):
    def create_pending(self, *, user, lineitems=None, **kwargs):
        """
        Create an unpaid payment instance with all line items for the given
        user that have not been bound to a payment instance yet.

        Returns ``None`` if there are no unbound line items for the given user.
        """
        with transaction.atomic():
            items = user.user_lineitems.unbound()
            if lineitems is not None:
                items = items.filter(pk__in=[i.pk for i in lineitems])
            if not len(items):
                return None

            payment = self.create(
                user=user, amount=sum((item.amount for item in items), 0), **kwargs
            )
            items.update(payment=payment)
            return payment


class Payment(AbstractPayment):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="user_payments",
        verbose_name=_("user"),
    )

    objects = PaymentManager.from_queryset(PaymentQuerySet)()

    def __str__(self):
        if self.charged_at:
            return gettext("Payment of %s") % self.amount
        return gettext("Pending payment of %s") % self.amount

    def save(self, *args, **kwargs):
        if not self.email:
            self.email = self.user.email
        super().save(*args, **kwargs)

    save.alters_data = True

    def cancel_pending(self):
        assert not self.charged_at
        self.lineitems.update(payment=None)
        self.delete()

    cancel_pending.alters_data = True

    def undo(self):
        assert self.charged_at
        self.charged_at = None
        self.save()
        # Unbind line items after post_save signal handling
        self.lineitems.update(payment=None)

    undo.alters_data = True

    @property
    def description(self):
        return "Payment of %s by %s: %s" % (
            self.amount,
            self.email,
            ", ".join(str(item) for item in self.lineitems.all()),
        )


class LineItemQuerySet(models.QuerySet):
    def unbound(self):
        return self.filter(payment__isnull=True)

    def unpaid(self):
        return self.filter(
            Q(payment__isnull=True) | Q(payment__charged_at__isnull=True)
        )


class LineItem(models.Model):
    """
    Individual line items may be created directly using the manager method. A
    minimal example follows::

        @login_required
        def some_view(request):
            # This request costs 5 cents!
            LineItem.objects.create(
                user=request.user,
                amount=Decimal('0.05'),
                title='Request to some_view',
            )
            # Rest of view

    If you already have a payment instance at hand you may pass it as well.
    """

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="user_lineitems",
        verbose_name=_("user"),
    )
    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    title = models.CharField(_("title"), max_length=200)
    amount = models.DecimalField(_("amount"), max_digits=10, decimal_places=2)

    payment = models.ForeignKey(
        Payment,
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        related_name="lineitems",
        verbose_name=_("payment"),
    )

    objects = LineItemQuerySet.as_manager()

    class Meta:
        ordering = ["-created_at"]
        verbose_name = _("line item")
        verbose_name_plural = _("line items")

    def __str__(self):
        return self.title



# Generated by Django 2.0.5 on 2018-05-21 19:26

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="Customer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="updated at"),
                ),
                (
                    "customer_id",
                    models.CharField(
                        max_length=50, unique=True, verbose_name="customer ID"
                    ),
                ),
                (
                    "customer_data",
                    models.TextField(blank=True, verbose_name="customer data"),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="stripe_customer",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={"verbose_name": "customer", "verbose_name_plural": "customers"},
        )
    ]



# Generated by Django 2.0.5 on 2018-05-21 05:11

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("user_payments", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Subscription",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "code",
                    models.CharField(
                        help_text="Codes must be unique per user. Allows identifying the subscription.",
                        max_length=20,
                        verbose_name="code",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                ("starts_on", models.DateField(verbose_name="starts on")),
                (
                    "ends_on",
                    models.DateField(blank=True, null=True, verbose_name="ends on"),
                ),
                (
                    "periodicity",
                    models.CharField(
                        choices=[
                            ("yearly", "yearly"),
                            ("quarterly", "quarterly"),
                            ("monthly", "monthly"),
                            ("weekly", "weekly"),
                            ("manually", "manually"),
                        ],
                        max_length=20,
                        verbose_name="periodicity",
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2, max_digits=10, verbose_name="amount"
                    ),
                ),
                (
                    "renew_automatically",
                    models.BooleanField(
                        default=True, verbose_name="renew automatically"
                    ),
                ),
                ("paid_until", models.DateField(blank=True, verbose_name="paid until")),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="user_subscriptions",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "subscription",
                "verbose_name_plural": "subscriptions",
            },
        ),
        migrations.CreateModel(
            name="SubscriptionPeriod",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("starts_on", models.DateField(verbose_name="starts on")),
                ("ends_on", models.DateField(verbose_name="ends on")),
                (
                    "line_item",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="user_payments.LineItem",
                        verbose_name="line item",
                    ),
                ),
                (
                    "subscription",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="periods",
                        to="user_subscriptions.Subscription",
                        verbose_name="subscription",
                    ),
                ),
            ],
            options={
                "verbose_name": "subscription period",
                "verbose_name_plural": "subscription periods",
            },
        ),
        migrations.AlterUniqueTogether(
            name="subscriptionperiod", unique_together={("subscription", "starts_on")}
        ),
        migrations.AlterUniqueTogether(
            name="subscription", unique_together={("user", "code")}
        ),
    ]



import hashlib
import json

from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

import stripe


class CustomerManager(models.Manager):
    def with_token(self, *, user, token):
        """
        Add or replace a credit card for a given user
        """

        try:
            user.stripe_customer
        except Customer.DoesNotExist:
            return self._create_with_token(user, token)
        else:
            return self._update_token(user, token)

    def _create_with_token(self, user, token):
        obj = stripe.Customer.create(
            email=user.email,
            source=token,
            expand=["default_source"],
            idempotency_key="create-with-token-%s"
            % (hashlib.sha1(token.encode("utf-8")).hexdigest(),),
        )
        customer, created = self.update_or_create(
            user=user,
            defaults={"customer_id": obj.id, "customer_data": json.dumps(obj)},
        )
        customer._customer_data_cache = obj
        return customer

    def _update_token(self, user, token):
        obj = stripe.Customer.retrieve(user.stripe_customer.customer_id)
        obj.source = token
        obj.save()
        user.stripe_customer.refresh()
        return user.stripe_customer


class Customer(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        verbose_name=_("user"),
        related_name="stripe_customer",
    )
    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    updated_at = models.DateTimeField(_("updated at"), auto_now=True)
    customer_id = models.CharField(_("customer ID"), max_length=50, unique=True)
    customer_data = models.TextField(_("customer data"), blank=True)

    objects = CustomerManager()

    class Meta:
        verbose_name = _("customer")
        verbose_name_plural = _("customers")

    def __str__(self):
        return "%s%s" % (self.customer_id[:10], "*" * (len(self.customer_id) - 10))

    def save(self, *args, **kwargs):
        if hasattr(self, "_customer_data_cache"):
            self.customer_data = json.dumps(self._customer_data_cache)
        if not self.customer_data:
            self.refresh(save=False)
        super().save(*args, **kwargs)

    save.alters_data = True

    @property
    def customer(self):
        """
        Return the parsed version of the JSON blob or ``None`` if there is no
        data around to be parsed.

        After calling ``Customer.refresh()`` this property even returns the
        objects returned by the Stripe library as they come.

        Does NOT work with ``instance.refresh_from_db()`` -- you have to fetch
        a completely new object from the database.
        """
        if not hasattr(self, "_customer_data_cache"):
            self._customer_data_cache = json.loads(self.customer_data or "{}")
        return self._customer_data_cache

    @customer.setter
    def customer(self, value):
        self._customer_data_cache = value
        self.customer_data = json.dumps(self._customer_data_cache)

    def refresh(self, save=True):
        self.customer = stripe.Customer.retrieve(
            self.customer_id, expand=["default_source"]
        )
        if save:
            self.save()
